/* wmudmount
 * Copyright © 2010  Brad Jorsch <anomie@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <math.h>

#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include "conf.h"
#include "die.h"
#include "dock.h"
#include "notify.h"
#include "fsm.h"
#include "chars.h"
#include "menu.h"
#include "misc.h"

#include "master.xpm"

/* Globals */
static GdkDisplay *display;
static GtkWidget *dockwin, *iconwin;
static GdkPixmap *pixmap;
static GdkBitmap *mask;
static GdkGC *pixmap_gc;
static GtkIconTheme *theme;

#define LEFT_BUTTON 1
#define RIGHT_BUTTON 2
#define MOUNT_BUTTON 3
#define EJECT_BUTTON 4
static int button_down = 0;

static int cur_index = -1;
static udisks_volume *fsm_vols[7];
static udisks_volume *menu_volume = NULL;
static int hover_row = -1;

gint return_to_fsm_timeout = 60;
gboolean skip_root_only_filesystems = FALSE;

/* Utility funcs */
static void draw_button(GdkWindow *w, int dx, int dy, int img, gboolean pressed){
    gdk_draw_drawable(w, pixmap_gc, pixmap, pressed?108:96,52, dx,dy, 12,12);
    gdk_draw_drawable(w, pixmap_gc, pixmap, 120+10*(img%2),10*(img/2), dx+1,dy+1, 10,10);
}

static void draw_fsm(GdkWindow *w){
    udisks_volume *v;
    int n = 0;
    double pcts[7];
    for(int i=0; i<7; i++){
        fsm_vols[i] = NULL;
        pcts[i] = -1;
    }
    for(int i=0; i<num_volumes; i++){
        v = Volumes[i];
        if(v->type != UDISKS_FILESYSTEM || !v->d.fs.mountpoint || v->d.fs.total_size==0) continue;
        double pct = (v->d.fs.total_size - v->d.fs.free_size) / (double)v->d.fs.total_size;
        fsm_vols[n] = v;
        pcts[n] = pct;
        for(int j=n-1; j>=0; j--){
            if(pct <= pcts[j]) break;
            fsm_vols[j+1] = fsm_vols[j];
            pcts[j+1] = pcts[j];
            fsm_vols[j] = v;
            pcts[j] = pct;
        }
        if(n < 6) n++;
    }
    if(!n) return;

    static const GdkRectangle rect = { .x = 4, .y = 4, .width = 56, .height = 40 };
    gdk_gc_set_clip_rectangle(pixmap_gc, &rect);
    for(int i=0; i<n; i++){
        const char *name = fsm_vols[i]->label;
        if(!name) name = fsm_vols[i]->d.fs.mountpoint;
        int ww = measure_string(name);
        if(ww < 25) ww = 25;
        draw_string(w, pixmap_gc, name, 29-ww, 6+i*6, (i==hover_row)?2:0);
        int thresh = (pcts[i] > BLINK_THRESHOLD);
        ww = (fsm_blink && thresh)?0:(int)round(30 * pcts[i]);
        if(ww) gdk_draw_drawable(w, pixmap_gc, pixmap, 64,thresh?57:52, 30,6+i*6, ww,5);
    }
}

static GdkPixbuf *get_icon(udisks_volume *v){
    if(!v->icon) return NULL;

    GdkPixbuf *icon = NULL;
    GtkIconInfo *info = gtk_icon_theme_lookup_by_gicon(theme, v->icon, 24, GTK_ICON_LOOKUP_USE_BUILTIN);
    if(info){
        GError *err = NULL;
        icon = gtk_icon_info_load_icon(info, &err);
        if(!icon){
            char *c = g_icon_to_string(v->icon);
            warn(DEBUG_ERROR, "Couldn't load icon pixmap for %s: %s", c, err?err->message:"Unknown error");
            g_free(c);
            if(err) g_error_free(err);
        }
        gtk_icon_info_free(info);
    } else {
        char *c = g_icon_to_string(v->icon);
        warn(DEBUG_ERROR, "Couldn't load icon for %s", c);
        g_free(c);
    }
    return icon;
}

static udisks_volume *get_volume_at(int x, int y, GdkRectangle *rect){
    udisks_volume *v;
    if(cur_index == -1){
        int i = y/6 - 1;
        if(i < 0 || i > 5) return NULL;
        v = fsm_vols[i];
        if(!v) return NULL;
        if(rect){
            rect->x = 4; rect->y = 6 + i*6; rect->width = 56; rect->height = 6;
        }
        return v;
    } else {
        v = Volumes[cur_index];
        if(rect){
            rect->x = 4; rect->y = 4; rect->width = 40; rect->height = 56;
        }
        return v;
    }
}

static int return_to_fsm_counter = -1;
static gboolean return_to_fsm_cb(gpointer p){
    if(return_to_fsm_counter < 0) return TRUE;
    if(--return_to_fsm_counter >= 0) return TRUE;
    if(cur_index != -1){
        cur_index = -1;
        redraw_dock();
    }
    return TRUE;
}

/* All this mess is necessary to avoid firing the single-click command twice on
 * a double-click. Basically, schedule a timer for after the double-click
 * timeout, and have any button presses cancel the timer. */
static guint sgl_click_id = 0;
static udisks_volume *sgl_click_volume = NULL; // So we can udisks_unref it
static gboolean single_click_cb(gpointer p){
    udisks_volume *v = sgl_click_volume;
    sgl_click_volume = NULL;
    sgl_click_id = 0;
    if(!v) return FALSE;
    if(user_command_enabled(v,&menu_commands[sglclick_command_idx])){
        warn(DEBUG_DEBUG, "Executing user command \"%s\"", menu_commands[sglclick_command_idx].title);
        exec_command(v, &menu_commands[sglclick_command_idx]);
    } else {
        warn(DEBUG_DEBUG, "User command \"%s\" preconditions not met", menu_commands[sglclick_command_idx].title);
    }
    udisks_unref(v);
    return FALSE;
}
static void cancel_sgl_click(){
    if(sgl_click_volume){
        udisks_unref(sgl_click_volume);
        sgl_click_volume = NULL;
    }
    if(sgl_click_id){
        g_source_remove(sgl_click_id);
        sgl_click_id = 0;
    }
}
static void schedule_sgl_click(udisks_volume *v){
    if(dblclick_command_idx<0){
        if(!v) return;
        if(user_command_enabled(v,&menu_commands[sglclick_command_idx])){
            warn(DEBUG_DEBUG, "Executing user command \"%s\"", menu_commands[sglclick_command_idx].title);
            exec_command(v, &menu_commands[sglclick_command_idx]);
        } else {
            warn(DEBUG_DEBUG, "User command \"%s\" preconditions not met", menu_commands[sglclick_command_idx].title);
        }
    } else {
        cancel_sgl_click();
        if(v){
            udisks_ref(v);
            sgl_click_volume = v;
            guint timeout = 400;
            g_object_get(G_OBJECT(gtk_settings_get_default()), "gtk-double-click-time", &timeout, NULL);
            warn(DEBUG_DEBUG, "Setting timeout of %d ms for single-click command", timeout);
            sgl_click_id = g_timeout_add(timeout, single_click_cb, NULL);
        }
    }
}

/* GTK event handlers */

static char *concat_printf(char *buf, char *fmt, ...) __attribute__ ((format (printf, 2, 3)));
static char *concat_printf(char *buf, char *format, ...){
    va_list ap;
    va_start(ap, format);

    char *ret;
    char *tmp = g_markup_vprintf_escaped(format, ap);
    if(!buf){
        ret = tmp;
    } else {
        ret = g_strjoin("\n", buf, tmp, NULL);
        g_free(buf);
        g_free(tmp);
    }

    va_end(ap);
    return ret;
}

static gboolean query_tooltip_handler(GtkWidget *widget, gint x, gint y, gboolean keyboard_mode, GtkTooltip *tt, gpointer data){
    GdkRectangle rect;
    udisks_volume *v = get_volume_at(x,y,&rect);
    if(!v) return FALSE;
    GdkPixbuf *icon = get_icon(v);
    gtk_tooltip_set_icon(tt, icon);
    g_object_unref(icon);

    if(y >= 4 && y <= 44 && x >= 4 && x <= 60){
        gtk_tooltip_set_tip_area(tt, &rect);
        char *c = NULL;
        if(v->dev) c = concat_printf(c, "<b>Device:</b> %s", v->dev);
        if(v->label) c = concat_printf(c, "<b>Label:</b> %s", v->label);
        if(v->desc) c = concat_printf(c, "<b>Description:</b> %s", v->desc);
        if(v->uuid) c = concat_printf(c, "<b>UUID:</b> %s", v->uuid);
        switch(v->type){
          case UDISKS_FILESYSTEM:
            if(v->d.fs.mountpoint) c=concat_printf(c, "<b>Mountpoint:</b> %s", v->d.fs.mountpoint);
            if(v->d.fs.total_size!=0){
                double pct = (v->d.fs.total_size - v->d.fs.free_size) / (double)v->d.fs.total_size;
                char *t = format_byte_size(v->d.fs.total_size);
                char *u = format_byte_size(v->d.fs.total_size - v->d.fs.free_size);
                char *f = format_byte_size(v->d.fs.free_size);
                if(pct >= BLINK_THRESHOLD){
                    c=concat_printf(c, "<span foreground=\"red\"><b>Space Used:</b> %s of %s (%.1f%%), %s free</span>", u, t, 100.0*pct, f);
                } else {
                    c=concat_printf(c, "<b>Space Used:</b> %s of %s (%.1f%%), %s free", u, t, 100.0*pct, f);
                }
                g_free(t);
                g_free(u);
                g_free(f);
            }
            break;
          case UDISKS_LUKS:
            break;
        }
        if(!c) return FALSE;
        if(udisks_daemon) c = concat_printf(c, "<b>Daemon:</b> %s", udisks_daemon);
        gtk_tooltip_set_markup(tt, c);
        g_free(c);
        return TRUE;
    } else if(y >= 48 && y <= 59){
        rect.y = 48; rect.width=12; rect.height=12;
        if(x >= 4 && x <= 15){
            rect.x = 4; gtk_tooltip_set_tip_area(tt, &rect);
            switch(v->type){
              case UDISKS_FILESYSTEM:
                if(v->d.fs.mountpoint){
                    gtk_tooltip_set_text(tt, "Unmount the filesystem");
                } else if(udisks_can_mount(v)){
                    gtk_tooltip_set_text(tt, "Mount the filesystem");
                } else {
                    return FALSE;
                }
                return TRUE;
              case UDISKS_LUKS:
                gtk_tooltip_set_text(tt, "Unlock the device");
                return TRUE;
            }
        } else if(x >= 16 && x <= 25){
            rect.x = 16; gtk_tooltip_set_tip_area(tt, &rect);
            if(udisks_can_lock(v)){
                gtk_tooltip_set_text(tt, "Lock the device");
                return TRUE;
            } else if(udisks_can_eject(v)){
                gtk_tooltip_set_text(tt, "Eject the media from the drive");
                return TRUE;
            } else if(udisks_can_detach(v)){
                gtk_tooltip_set_text(tt, "Detach the drive");
                return TRUE;
            }
        }
    }

    return FALSE;
}

static gboolean expose_handler(GtkWidget *widget, GdkEventExpose *ev, gpointer d){
    GdkWindow *w = widget->window;

    warn(DEBUG_DEBUG, "Redrawing %p (%d,%d+%d+%d)", w, ev->area.x, ev->area.y, ev->area.width, ev->area.height);
    gdk_gc_set_clip_region(pixmap_gc, ev->region);

    gdk_draw_drawable(w, pixmap_gc, pixmap, 0,0, 0,0, 64,64);

    if(!udisks_daemon){
        int ww = measure_string("Udisks daemon");
        draw_string(w, pixmap_gc, "Udisks daemon", 32-ww/2, 24-char_height-1, 3);
        ww = measure_string("is unavailable");
        draw_string(w, pixmap_gc, "is unavailable", 32-ww/2, 25, 3);
        return FALSE;
    }

    gboolean show_buttons = FALSE;
    if(skip_root_only_filesystems){
        for(int i=0; !show_buttons && i<num_volumes; i++){
            if(!Volumes[i]->root_only) show_buttons = TRUE;
        }
    } else {
        show_buttons = (num_volumes > 0);
    }
    if(show_buttons){
        draw_button(w, 36,48, 0, button_down == LEFT_BUTTON);
        draw_button(w, 48,48, 1, button_down == RIGHT_BUTTON);
    }

    if(cur_index == -1){
        draw_fsm(w);
        return FALSE;
    }

    udisks_volume *v = Volumes[cur_index];
    int highlight = 0;
    const char *bottom_text = NULL;
    int bottom_color = 0;
    int bar_width = 0;
    int bar_color = 0;
    switch(v->type){
      case UDISKS_FILESYSTEM:
        if(v->d.fs.mountpoint){
            highlight = 1;
            bottom_text = v->d.fs.mountpoint;
            bottom_color = 0;
            double pct = (v->d.fs.total_size - v->d.fs.free_size) / (double)v->d.fs.total_size;
            bar_color = (pct > BLINK_THRESHOLD)?1:0;
            bar_width = (fsm_blink && bar_color)?0:(int)round(56 * pct);
            draw_button(w, 4,48, 5, button_down == MOUNT_BUTTON);
        } else {
            highlight = 0;
            bottom_text = "<unmounted>";
            bottom_color = 1;
            bar_width = 0;
            draw_button(w, 4,48, 4, button_down == MOUNT_BUTTON);
        }
        if(v->pending_actions) highlight = 2;
        break;
      case UDISKS_LUKS:
        bottom_text = "LUKS device";
        if(v->pending_actions) highlight = 2;
        draw_button(w, 4,48, 6, button_down == MOUNT_BUTTON);
        break;
    }
    if(udisks_can_lock(v)){
        draw_button(w, 16,48, 7, button_down == EJECT_BUTTON);
    } else if(udisks_can_eject(v)){
        draw_button(w, 16,48, 2, button_down == EJECT_BUTTON);
    } else if(udisks_can_detach(v)){
        draw_button(w, 16,48, 3, button_down == EJECT_BUTTON);
    }

    static const GdkRectangle rect = { .x = 4, .y = 4, .width = 56, .height = 40 };
    GdkRegion *region = gdk_region_rectangle(&rect);
    gdk_region_intersect(region, ev->region);
    if(!gdk_region_empty(region)){
        gdk_gc_set_clip_region(pixmap_gc, region);

        if(highlight) gdk_draw_drawable(w, pixmap_gc, pixmap, 64,(highlight-1)*24, 4,11, 56,24);

        const char *s = udisks_get_display_name(v);
        int x = 4;
        int rj = udisks_display_name_should_right_justify(v);
        if(rj){
            x = 60-measure_string(s);
            if(x > 4) x = 4;
        }
        draw_string(w, pixmap_gc, s, x, 5, rj?1:0);

        if(bar_width) gdk_draw_drawable(w, pixmap_gc, pixmap, 64,48+bar_color*2, 4,35, bar_width,2);
        if(bottom_text) draw_string(w, pixmap_gc, bottom_text, 4, 38, bottom_color);

        GdkPixbuf *icon = get_icon(v);
        if(icon){
            gdk_draw_pixbuf(w, pixmap_gc, icon, 0,0, 20,11, 24,24, GDK_RGB_DITHER_NONE,0,0);
            g_object_unref(icon);
        }

        gdk_region_destroy(region);
    }

    return FALSE;
}

static gboolean style_set_handler(GtkWidget *w, GtkStyle *s, gpointer d){
    static const GdkRectangle rect = { .x = 20, .y = 10, .width = 24, .height = 24 };
    warn(DEBUG_DEBUG, "Theme changed");
    if(w->window) gdk_window_invalidate_rect(w->window, &rect, FALSE);
    return FALSE;
}

void dock_action_callback(udisks_volume *v, gpointer ret, gpointer data, gboolean failed, GError *err){
    struct notify_actions *actions = dock_actions_for_volume(v);
    if(failed){
        notify(DEBUG_ERROR, "device.error", FALSE, v->icon, actions, "%s failed", "%s of %s failed: %s", data, udisks_get_display_name(v), err?err->message:"<unknown error>");
    } else if(err){
        int ct = 0;
        for(char *p=err->message; *p; p++, ct++){
            if(*p=='%') ct++;
        }
        char buf[ct];
        for(char *p=err->message, *q=buf; *p; *q++=*p++){
            if(*p=='%') *q++='%';
        }
        notify(DEBUG_INFO, "device", TRUE, v->icon, actions, "%s succeeded", err->message, data);
    } else {
        notify(DEBUG_INFO, "device", TRUE, v->icon, actions, "%s succeeded", "%s of %s succeeded", data, udisks_get_display_name(v));
    }
    g_slice_free(struct notify_actions, actions);
    redraw_dock();
}
void dock_lock_callback(udisks_volume *v, gpointer ret, gpointer data, gboolean failed, GError *err){
    if(!failed){
        for(int i=0; i<num_volumes; i++){
            if(Volumes[i] == (udisks_volume *)ret && !(skip_root_only_filesystems && Volumes[i]->root_only)) cur_index = i;
        }
    }
    dock_action_callback(v, ret, data, failed, err);
}

static gboolean click_handler(GtkWidget *w, GdkEventButton *ev, gpointer d){
    int x=ev->x, y=ev->y;
    udisks_volume *v;

    if(ev->button == 2){
        if(ev->type==GDK_BUTTON_PRESS && cur_index!=-1){
            cur_index = -1;
            redraw_dock();
        }
    } else if(ev->button == 3){
        v = get_volume_at(x, y, NULL);
        if(update_menu(v, cur_index==-1)){
            menu_volume = v;
            show_menu(w, ev);
        }
    }
    if(ev->button != 1) return FALSE;

    int button = 0;
    if(y >= 48 && y <= 59){
        if(x >= 4 && x <= 15) button = MOUNT_BUTTON;
        if(x >= 16 && x <= 25) button = EJECT_BUTTON;
        if(x >= 36 && x <= 47) button = LEFT_BUTTON;
        if(x >= 48 && x <= 59) button = RIGHT_BUTTON;
    }

    switch(ev->type){
      case GDK_BUTTON_PRESS:
        warn(DEBUG_DEBUG, "Button %d press at %d,%d", ev->button, x, y);
        if(button_down != button) redraw_dock();
        button_down = button;
        cancel_sgl_click();
        if(sglclick_command_idx>=0 && y >= 4 && y <= 44 && x >= 4 && x <= 60){
            schedule_sgl_click(get_volume_at(x, y, NULL));
        }
        break;

      case GDK_BUTTON_RELEASE:
        warn(DEBUG_DEBUG, "Button %d release at %d,%d", ev->button, x, y);
        if(button_down == button) switch(button){
          case MOUNT_BUTTON:
            mount_button_click();
            break;
          case EJECT_BUTTON:
            eject_button_click();
            break;
          case LEFT_BUTTON:
            left_button_click();
            break;
          case RIGHT_BUTTON:
            right_button_click();
            break;
        }
        if(button_down) redraw_dock();
        button_down = 0;
        break;

      case GDK_2BUTTON_PRESS:
        warn(DEBUG_DEBUG, "Button %d double-click at %d,%d", ev->button, x, y);
        cancel_sgl_click();
        if(dblclick_command_idx>=0 && y >= 4 && y <= 44 && x >= 4 && x <= 60){
            v = get_volume_at(x, y, NULL);
            if(!v) break;
            if(user_command_enabled(v,&menu_commands[dblclick_command_idx])){
                warn(DEBUG_DEBUG, "Executing user command \"%s\"", menu_commands[dblclick_command_idx].title);
                exec_command(v, &menu_commands[dblclick_command_idx]);
            } else {
                warn(DEBUG_DEBUG, "User command \"%s\" preconditions not met", menu_commands[dblclick_command_idx].title);
            }
        }
        break;

      default:
        break;
    }

    return FALSE;
}

void mount_button_click(){
    if(cur_index < 0) return;
    dock_action_mount(Volumes[cur_index]);
}

void dock_action_mount(udisks_volume *v){
    switch(v->type){
      case UDISKS_FILESYSTEM:
        if(v->d.fs.mountpoint){
            udisks_unmount(v, dock_action_callback, "Unmount");
        } else if(udisks_can_mount(v)){
            udisks_mount(v, dock_action_callback, "Mount");
        }
        break;
      case UDISKS_LUKS:
        udisks_unlock(v, dock_lock_callback, "Unlock");
        break;
    }
}

void eject_button_click(){
    if(cur_index < 0) return;
    dock_action_eject(Volumes[cur_index]);
}

void dock_action_eject(udisks_volume *v){
    if(udisks_can_lock(v)){
        udisks_lock(v, dock_lock_callback, "Lock");
    } else if(udisks_can_eject(v)){
        udisks_eject(v, dock_action_callback, "Eject");
    } else if(udisks_can_detach(v)){
        udisks_detach(v, dock_action_callback, "Detach");
    }
}

void left_button_click(){
    close_menu();
    do {
        if(++cur_index >= num_volumes) cur_index=-1;
    } while(cur_index>=0 && skip_root_only_filesystems && Volumes[cur_index]->root_only);
}

void right_button_click(){
    close_menu();
    if(cur_index < 0) cur_index=num_volumes;
    do {
        --cur_index;
    } while(cur_index>=0 && skip_root_only_filesystems && Volumes[cur_index]->root_only);
}

void dock_action_goto(udisks_volume *v){
    for(int i=0; i<num_volumes; i++){
        if(Volumes[i] == v && !(skip_root_only_filesystems && Volumes[i]->root_only) && cur_index != i){
            cur_index = i;
            redraw_dock();
        }
    }
}

struct notify_actions *dock_actions_for_volume(udisks_volume *v){
    struct notify_actions *ret = g_slice_new(struct notify_actions);
    ret->volume = v;
    ret->mount_text = NULL;
    switch(v->type){
      case UDISKS_FILESYSTEM:
        if(v->d.fs.mountpoint){
            ret->mount_text = "Unmount";
        } else if(udisks_can_mount(v)){
            ret->mount_text = "Mount";
        }
        break;
      case UDISKS_LUKS:
        ret->mount_text = "Unlock";
        break;
    }
    ret->eject_text = NULL;
    if(udisks_can_lock(v)){
        ret->eject_text = "Lock";
    } else if(udisks_can_eject(v)){
        ret->eject_text = "Eject";
    } else if(udisks_can_detach(v)){
        ret->eject_text = "Detach";
    }
    return ret;
}

static gboolean scroll_handler(GtkWidget *w, GdkEventScroll *ev, gpointer d){
    int x=ev->x, y=ev->y;

    if(ev->type != GDK_SCROLL) return FALSE;
    warn(DEBUG_DEBUG, "Scroll %d at %d,%d", ev->direction, x, y);

    switch(ev->direction){
      case GDK_SCROLL_UP:
      case GDK_SCROLL_LEFT:
        close_menu();
        do {
            if(++cur_index >= num_volumes) cur_index=-1;
        } while(cur_index>=0 && skip_root_only_filesystems && Volumes[cur_index]->root_only);
        redraw_dock();
        break;
      case GDK_SCROLL_DOWN:
      case GDK_SCROLL_RIGHT:
        close_menu();
        if(cur_index < 0) cur_index=num_volumes;
        do {
            --cur_index;
        } while(cur_index>=0 && skip_root_only_filesystems && Volumes[cur_index]->root_only);
        redraw_dock();
        break;
    }

    return FALSE;
}

static gboolean motion_handler(GtkWidget *w, GdkEventMotion *ev, gpointer d){
    int old_hover_row = hover_row;
    hover_row = ((int)ev->y)/6 - 1;
    if(cur_index == -1 && old_hover_row != hover_row) redraw_dock();
    return FALSE;
}

static gboolean enter_handler(GtkWidget *w, GdkEventCrossing *ev, gpointer d){
    return_to_fsm_counter = -1;
    int old_hover_row = hover_row;
    hover_row = ((int)ev->y)/6 - 1;
    if(cur_index == -1 && old_hover_row != hover_row) redraw_dock();
    return FALSE;
}

static gboolean leave_handler(GtkWidget *w, GdkEventCrossing *ev, gpointer d){
    return_to_fsm_counter = return_to_fsm_timeout;
    if(cur_index == -1 && hover_row >= 0) redraw_dock();
    hover_row = -1;
    return FALSE;
}


/* Functions */

void redraw_dock(void){
    gdk_window_invalidate_rect(dockwin->window, NULL, FALSE);
    gdk_window_invalidate_rect(iconwin->window, NULL, FALSE);
    gtk_tooltip_trigger_tooltip_query(display);
}

void create_dock_icon(void){
    warn(DEBUG_DEBUG, "Loading theme");
    theme = gtk_icon_theme_get_default();
    display = gdk_display_get_default();

    warn(DEBUG_INFO, "Creating windows");
    dockwin=gtk_window_new(GTK_WINDOW_TOPLEVEL);
    if(dockwin==NULL) die("Couldn't create window");
    iconwin=gtk_window_new(GTK_WINDOW_POPUP);
    if(iconwin==NULL) die("Couldn't create window");
    gtk_window_set_wmclass(GTK_WINDOW(dockwin), g_get_prgname(), "DockApp");
    gtk_window_set_wmclass(GTK_WINDOW(iconwin), g_get_prgname(), "DockApp");
    gtk_widget_set_size_request(dockwin, 64, 64);
    gtk_widget_set_size_request(iconwin, 64, 64);
    gtk_widget_set_app_paintable(dockwin, TRUE);
    gtk_widget_set_app_paintable(iconwin, TRUE);
    gtk_widget_add_events(dockwin, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_EXPOSURE_MASK | GDK_SCROLL_MASK | GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK | GDK_POINTER_MOTION_HINT_MASK);
    gtk_widget_add_events(iconwin, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_EXPOSURE_MASK | GDK_SCROLL_MASK | GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK | GDK_POINTER_MOTION_HINT_MASK);
    g_object_set(dockwin, "has-tooltip", TRUE, NULL);
    g_object_set(iconwin, "has-tooltip", TRUE, NULL);
    g_signal_connect(G_OBJECT(dockwin), "expose-event", G_CALLBACK(expose_handler), NULL);
    g_signal_connect(G_OBJECT(iconwin), "expose-event", G_CALLBACK(expose_handler), NULL);
    g_signal_connect(G_OBJECT(dockwin), "button-press-event", G_CALLBACK(click_handler), NULL);
    g_signal_connect(G_OBJECT(iconwin), "button-press-event", G_CALLBACK(click_handler), NULL);
    g_signal_connect(G_OBJECT(dockwin), "button-release-event", G_CALLBACK(click_handler), NULL);
    g_signal_connect(G_OBJECT(iconwin), "button-release-event", G_CALLBACK(click_handler), NULL);
    g_signal_connect(G_OBJECT(dockwin), "scroll-event", G_CALLBACK(scroll_handler), NULL);
    g_signal_connect(G_OBJECT(iconwin), "scroll-event", G_CALLBACK(scroll_handler), NULL);
    g_signal_connect(G_OBJECT(dockwin), "enter-notify-event", G_CALLBACK(enter_handler), NULL);
    g_signal_connect(G_OBJECT(iconwin), "enter-notify-event", G_CALLBACK(enter_handler), NULL);
    g_signal_connect(G_OBJECT(dockwin), "leave-notify-event", G_CALLBACK(leave_handler), NULL);
    g_signal_connect(G_OBJECT(iconwin), "leave-notify-event", G_CALLBACK(leave_handler), NULL);
    g_signal_connect(G_OBJECT(dockwin), "motion-notify-event", G_CALLBACK(motion_handler), NULL);
    g_signal_connect(G_OBJECT(iconwin), "motion-notify-event", G_CALLBACK(motion_handler), NULL);
    g_signal_connect(G_OBJECT(dockwin), "style-set", G_CALLBACK(style_set_handler), NULL);
    g_signal_connect(G_OBJECT(iconwin), "style-set", G_CALLBACK(style_set_handler), NULL);
    g_signal_connect(G_OBJECT(dockwin), "query-tooltip", G_CALLBACK(query_tooltip_handler), NULL);
    g_signal_connect(G_OBJECT(iconwin), "query-tooltip", G_CALLBACK(query_tooltip_handler), NULL);
    g_signal_connect(G_OBJECT(dockwin), "destroy", G_CALLBACK(gtk_main_quit), NULL);
    g_signal_connect(G_OBJECT(iconwin), "destroy", G_CALLBACK(gtk_main_quit), NULL);

    warn(DEBUG_DEBUG, "Realizing windows");
    gtk_widget_realize(dockwin);
    gtk_widget_realize(iconwin);
    gdk_window_set_icon(dockwin->window, iconwin->window, NULL, NULL);

    warn(DEBUG_DEBUG, "Loading pixmaps");
    pixmap=gdk_pixmap_create_from_xpm_d(iconwin->window, &mask, NULL, master_xpm);
    if(!pixmap) die("Could not load master pixmap");
    if(!init_chars(iconwin->window, 4,
                   0,0,0, 0x15,0x8F,0xBC,
                   0,0,0, 0x75,0x75,0x75,
                   0,0,0, 0x1c,0xc2,0xff,
                   0,0,0, 0xff,0x00,0x00
                  )) die("Could not load character pixmaps");
    pixmap_gc=gdk_gc_new(iconwin->window);

    gtk_widget_shape_combine_mask(dockwin, mask, 0, 0);
    gtk_widget_shape_combine_mask(iconwin, mask, 0, 0);

    gtk_widget_show(iconwin);

#ifdef GTK_WITHDRAWN_HACK
    // Gtk/Gdk gives us no way to specify intial_state = WithdrawnState,
    // because ICCCM doesn't list it in the little table for initial_state.
    // And it even overrides it just before mapping the window!
    // See https://bugzilla.gnome.org/show_bug.cgi?id=139322
    //
    // But there is a hack: an X window isn't actually mapped until its parent
    // is mapped. So if we reparent the troublesome window to an unmapped
    // window and call GTK's map function, we can then easily change the hints
    // before reparenting it back to the root window.
    Window p, dummy1, *dummy2;
    unsigned int dummy3;
    XQueryTree(GDK_DISPLAY(),GDK_WINDOW_XID(dockwin->window),&dummy1,&p,&dummy2,&dummy3);
    if(dummy2) XFree(dummy2);
    Window w = XCreateSimpleWindow(GDK_DISPLAY(),p,0,0,1,1,0,0,0);
    XReparentWindow(GDK_DISPLAY(),GDK_WINDOW_XID(dockwin->window),w,0,0);
    gtk_widget_show(dockwin);
    XWMHints *hints = XGetWMHints(GDK_DISPLAY(),GDK_WINDOW_XID(dockwin->window));
    hints->flags |= StateHint;
    hints->initial_state = WithdrawnState;
    XSetWMHints(GDK_DISPLAY(),GDK_WINDOW_XID(dockwin->window),hints);
    XFree(hints);
    XReparentWindow(GDK_DISPLAY(),GDK_WINDOW_XID(dockwin->window),p,0,0);
    XDestroyWindow(GDK_DISPLAY(),w);
#else
    gtk_widget_show(dockwin);
    gdk_window_withdraw(dockwin->window);
#endif
    warn(DEBUG_DEBUG, "Create windows dockwin=%p iconwin=%p", dockwin->window, iconwin->window);

    g_timeout_add_seconds(1, return_to_fsm_cb, NULL);
}

/* Device notifications */

void volume_added(udisks_volume *v, gboolean connecting){
    if(!connecting){
        struct notify_actions *actions = dock_actions_for_volume(v);
        notify(NOTIFY_INFO, "device.added", TRUE, v->icon, actions, "Device %s added", NULL, udisks_get_display_name(v));
        g_slice_free(struct notify_actions, actions);
    }
    if(cur_index == -1) redraw_dock();
}

void volume_updated(udisks_volume *v){
    if(cur_index >= 0 && Volumes[cur_index]==v){
        if(skip_root_only_filesystems && v->root_only){
            cur_index=-1;
            if(menu_volume == v) close_menu();
        }
        redraw_dock();
    } else if(cur_index == -1){
        redraw_dock();
    }
    if(menu_volume == v) update_menu(menu_volume, cur_index==-1);
}

void fsm_updated(void){
    if(cur_index == -1){
       redraw_dock();
    } else {
       udisks_volume *v=Volumes[cur_index];
       if(v->type == UDISKS_FILESYSTEM && v->d.fs.mountpoint && v->d.fs.total_size!=0) redraw_dock();
    }
}

void volume_removed(int idx, udisks_volume *v, gboolean disconnecting){
    if(!disconnecting){
        notify(NOTIFY_INFO, "device.removed", TRUE, v->icon, NULL, "Device %s removed", NULL, udisks_get_display_name(v));
    }
    if(cur_index == idx || cur_index == -1){
        cur_index = -1;
        if(menu_volume == v) close_menu();
        redraw_dock();
    } else if(cur_index > idx) cur_index--;
}

void udisks_connection_changed(const char *udisks_version){
    if(gtk_main_level()){
        notify(NOTIFY_INFO, NULL, TRUE, NULL, NULL, udisks_version?"Connected to %s":"Udisks connection lost", NULL, udisks_version);
        redraw_dock();
    }
}
